@@ -28,6 +28,7 @@
 #include <fuse.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <inttypes.h>
 #include <string.h>
@@ -40,6 +41,7 @@
 #include <dirent.h>
 #include <signal.h>
 #include <getopt.h>
+#include <expat.h>
 
 #include <obexftp/obexftp.h>
 #include <obexftp/client.h>
@@ -59,7 +61,7 @@ typedef struct data_buffer data_buffer_t;
 struct data_buffer {
 	size_t size;
 	char *data;
-	int write_mode; /* is this a write buffer? */
+	bool write_mode; /* is this a write buffer? */
 };
 
 
@@ -68,13 +70,26 @@ static int transport = 0;
 static char *source = NULL; // "00:11:22:33:44:55"; "hci0";
 static char *device = NULL; // "00:11:22:33:44:55"; "/dev/ttyS0";
 static int channel = -1;
+static char *root = NULL; // "E:";
+static int root_len = 0; // Length of 'root'.
 
-static int nonblock = 0;
+static bool nonblock = false;
+
+// Free/total space (in bytes) to report if can't get real info.
+static int report_space = 0;
 
 static char *mknod_dummy = NULL; /* bad coder, no cookies! */
 
 static int nodal = 0;
 
+static char* translate_path(const char* path) {
+	char* tpath = calloc(sizeof(char), root_len + strlen(path) + 1);
+	if (root)
+		strcpy(tpath, root);
+	strcpy(tpath + root_len, path);
+	DEBUG("TRANSLATED: %s\n", tpath);
+	return tpath;
+}
 
 static int cli_open()
 {
@@ -97,7 +112,11 @@ static int cli_open()
         for (retry = 0; retry < 3; retry++) {
 
                 /* Connect */
+#ifdef SIEMENS_S45
+                if (obexftp_connect_src (cli, source, device, channel, UUID_S45, sizeof(UUID_S45)) >= 0)
+#else
                 if (obexftp_connect_src (cli, source, device, channel, UUID_FBS, sizeof(UUID_FBS)) >= 0)
+#endif
                         return 0;
                 /* Still trying to connect */
 		sleep(1);
@@ -145,6 +164,7 @@ static void ofs_disconnect()
 
 static int ofs_getattr(const char *path, struct stat *stbuf)
 {
+	char *tpath;
 	stat_entry_t *st;
 	int res;
 
@@ -178,7 +198,9 @@ static int ofs_getattr(const char *path, struct stat *stbuf)
 	if(res < 0)
 		return res; /* errno */
 	
-	st = obexftp_stat(cli, path);
+	tpath = translate_path(path);
+	st = obexftp_stat(cli, tpath);
+	free(tpath);
 
 	ofs_disconnect();
 	
@@ -201,6 +223,7 @@ static int ofs_getattr(const char *path, struct stat *stbuf)
 
 static int ofs_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
 {
+	char *tpath;
 	DIR *dir;
 	stat_entry_t *ent;
 	int res;
@@ -209,9 +232,11 @@ static int ofs_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
 	if(res < 0)
 		return res; /* errno */
 
-	dir = obexftp_opendir(cli, path);
+	tpath = translate_path(path);
+	dir = obexftp_opendir(cli, tpath);
 	
 	if (!dir) {
+		free(tpath);
 		ofs_disconnect();
 		return -ENOENT;
 	}
@@ -225,6 +250,7 @@ static int ofs_getdir(const char *path, fuse_dirh_t h, fuse_dirfil_t filler)
 			break;
 	}
 	obexftp_closedir(dir);
+	free(tpath);
 
 	ofs_disconnect();
 
@@ -248,6 +274,7 @@ static int ofs_mknod(const char *path, mode_t UNUSED(mode), dev_t UNUSED(dev))
 
 static int ofs_mkdir(const char *path, mode_t UNUSED(mode))
 {
+	char *tpath;
 	int res;
 
 	if(!path || *path != '/')
@@ -257,7 +284,9 @@ static int ofs_mkdir(const char *path, mode_t UNUSED(mode))
 	if(res < 0)
 		return res; /* errno */
 
-	(void) obexftp_setpath(cli, path, 1);
+	tpath = translate_path(path);
+	(void) obexftp_setpath(cli, tpath, 1);
+	free(tpath);
 
 	ofs_disconnect();
 
@@ -266,6 +295,7 @@ static int ofs_mkdir(const char *path, mode_t UNUSED(mode))
 
 static int ofs_unlink(const char *path)
 {
+	char *tpath;
 	int res;
 
 	if(!path || *path != '/')
@@ -275,7 +305,9 @@ static int ofs_unlink(const char *path)
 	if(res < 0)
 		return res; /* errno */
 
-	(void) obexftp_del(cli, path);
+	tpath = translate_path(path);
+	(void) obexftp_del(cli, tpath);
+	free(tpath);
 
 	ofs_disconnect();
 
@@ -285,6 +317,7 @@ static int ofs_unlink(const char *path)
 
 static int ofs_rename(const char *from, const char *to)
 {
+	char *tfrom, *tto;
 	int res;
 
 	if(!from || *from != '/')
@@ -297,7 +330,11 @@ static int ofs_rename(const char *from, const char *to)
 	if(res < 0)
 		return res; /* errno */
 
-	(void) obexftp_rename(cli, from, to);
+	tfrom = translate_path(from);
+	tto = translate_path(to);
+	(void) obexftp_rename(cli, tfrom, tto);
+	free(tto);
+	free(tfrom);
 
 	ofs_disconnect();
 
@@ -326,6 +363,7 @@ static int ofs_open(const char *UNUSED(path), struct fuse_file_info *fi)
 
 static int ofs_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *UNUSED(fi))
 {
+	char *tpath;
 	data_buffer_t *wb;
 	int res = 0;
 	size_t actual;
@@ -340,7 +378,10 @@ static int ofs_read(const char *path, char *buf, size_t size, off_t offset, stru
 		if(res < 0)
 			return res; /* errno */
 
-		(void) obexftp_get(cli, NULL, path);
+		tpath = translate_path(path);
+		(void) obexftp_get(cli, NULL, tpath);
+		free(tpath);
+
 		wb->size = cli->buf_size;
 		wb->data = cli->buf_data; /* would be better to memcpy this */
 		//cli->buf_data = NULL; /* now the data is ours -- without copying */
@@ -378,7 +419,7 @@ static int ofs_write(const char *path, const char *buf, size_t size, off_t offse
 	if (!wb->data)
 		return -1;
 	wb->size = newsize;
-	wb->write_mode = 1;
+	wb->write_mode = true;
 
 	DEBUG("memcpy to %p (%p) from %p cnt %zu\n", wb->data + offset, wb->data, buf, size);
 	(void) memcpy(&wb->data[offset], buf, size);
@@ -389,6 +430,7 @@ static int ofs_write(const char *path, const char *buf, size_t size, off_t offse
 /* careful, this can be a read release or a write release */
 static int ofs_release(const char *path, struct fuse_file_info *fi)
 {
+	char *tpath;
 	data_buffer_t *wb;
 	int res;
 	
@@ -401,7 +443,9 @@ static int ofs_release(const char *path, struct fuse_file_info *fi)
 		if(res < 0)
 			return res; /* errno */
 
-		(void) obexftp_put_data(cli, (uint8_t *)wb->data, wb->size, path);
+		tpath = translate_path(path);
+		(void) obexftp_put_data(cli, (uint8_t *)wb->data, wb->size, tpath);
+		free(tpath);
 
 		ofs_disconnect();
 
@@ -412,47 +456,295 @@ static int ofs_release(const char *path, struct fuse_file_info *fi)
 	return 0;
 }
 
-#ifdef SIEMENS
-static int ofs_statfs(const char *UNUSED(label), struct statfs *st)
+struct parsing_data {
+	/* Input */
+	const char *path;
+
+	/* Intermediate states */
+	enum state_t {
+		PS_ROOT = 0,
+		PS_CAPABILITY,
+		PS_GENERAL,
+		PS_MEMORY,
+		PS_MEM_TYPE,
+		PS_LOCATION,
+		PS_FREE,
+		PS_USED,
+		PS_FILENLEN,
+
+		PS_FINISH,
+		PS_ERROR,
+	} state;               // State based on found XML tags
+	int depth;             // XML tag depth
+	char *mem_location;
+	unsigned long mem_used;
+	unsigned long mem_free;
+	unsigned long mem_namelen;
+
+	int path_match_len;    // Needed to find best-matching entry
+
+	/* Output */
+	const char *error;
+	struct statfs meminfo;
+};
+
+static void xml_element_start(void* user_data, const char* name, const char** UNUSED(attrs))
+{
+	struct parsing_data *data = (struct parsing_data*)user_data;
+
+	switch (data->state) {
+	case PS_ROOT:
+		if (data->depth == 0 && strcmp(name, "Capability") == 0) {
+			data->state = PS_CAPABILITY;
+		} else {
+			data->error = "Root element must be \"Capability\"";
+			data->state = PS_ERROR;
+		}
+		break;
+
+	case PS_CAPABILITY:
+		if (data->depth == 1 && strcmp(name, "General") == 0)
+			data->state = PS_GENERAL;
+		break;
+
+	case PS_GENERAL:
+		if (data->depth == 2 && strcmp(name, "Memory") == 0)
+			data->state = PS_MEMORY;
+		break;
+
+	case PS_MEMORY:
+		if (data->depth == 3) {
+			if (strcmp(name, "MemType") == 0)
+				data->state = PS_MEM_TYPE;
+			else if (strcmp(name, "Location") == 0)
+				data->state = PS_LOCATION;
+			else if (strcmp(name, "Free") == 0)
+				data->state = PS_FREE;
+			else if (strcmp(name, "Used") == 0)
+				data->state = PS_USED;
+			else if (strcmp(name, "FileNLen") == 0)
+				data->state = PS_FILENLEN;
+		}
+		break;
+
+	case PS_MEM_TYPE:
+	case PS_LOCATION:
+	case PS_FREE:
+	case PS_USED:
+	case PS_FILENLEN:
+		// Memory properties must be plain text.
+		data->error = "Memory properties must be plain text, no elements inside are allowed";
+		data->state = PS_ERROR;
+		break;
+
+	case PS_FINISH:
+	case PS_ERROR:
+		break;
+	}
+
+	++(data->depth);
+}
+
+static void xml_element_end(void* user_data, const char* name)
+{
+	struct parsing_data *data = (struct parsing_data*)user_data;
+	(data->depth)--;
+
+	switch (data->state) {
+	case PS_ROOT:
+		break;
+
+	case PS_CAPABILITY:
+		if (data->depth == 0 && strcmp(name, "Capability") == 0) {
+			data->error = "\"General\" section not found";
+			data->state = PS_ERROR;
+		}
+		break;
+
+	case PS_GENERAL:
+		if (data->depth == 1 && strcmp(name, "General") == 0)
+			data->state = PS_FINISH;
+		break;
+
+	case PS_MEMORY:
+		if (data->depth == 2 && strcmp(name, "Memory") == 0)
+		{
+
+			bool use_memory_entry = false;
+			ssize_t len = data->mem_location? strlen(data->mem_location): 0;
+
+			data->state = PS_GENERAL;
+
+			if (len == 0 && data->path_match_len < 0)
+			{
+				/* use memory entry in case that it contains no/empty
+				 * location entry and there was no previous location
+				 */
+				use_memory_entry = true;
+				data->path_match_len = 0;
+			}
+			else
+			{
+				/* Check if this memory entity is a better match */
+				if (data->path_match_len < len &&
+				    len <= (int)strlen(data->path) &&
+				    strncasecmp(data->path, data->mem_location, len) == 0 &&
+				    (data->path[len] == '/' || data->path[len] == 0))
+				{
+					use_memory_entry = true;
+					data->path_match_len = len;
+				}
+			}
+
+			if (use_memory_entry)
+			{
+				data->meminfo.f_bsize = 1;
+				data->meminfo.f_blocks = data->mem_used + data->mem_free;
+				data->meminfo.f_bfree = data->mem_free;
+				data->meminfo.f_bavail = data->mem_free;
+				data->meminfo.f_namelen = data->mem_namelen;
+			}
+
+			if (data->mem_location)
+				free(data->mem_location);
+		}
+		break;
+
+	case PS_MEM_TYPE:
+	case PS_LOCATION:
+	case PS_FREE:
+	case PS_USED:
+	case PS_FILENLEN:
+		data->state = PS_MEMORY;
+		break;
+
+	case PS_FINISH:
+	case PS_ERROR:
+		break;
+	}
+}
+
+static unsigned long parse_ulong(const char* str) {
+	long result;
+	errno = 0;
+	result = strtol(str, NULL, 10);
+	//check invalid string or invalid value
+	if (errno == 0 || result >= 0)
+		return (unsigned long)result;
+	return 0UL;
+}
+
+static void xml_character_data(void* user_data, const char* str, int len)
+{
+	struct parsing_data *data = (struct parsing_data*)user_data;
+	int i;
+	char *tmp = NULL;
+
+	switch (data->state) {
+	case PS_MEM_TYPE:
+		break;
+
+	case PS_LOCATION:
+		if (len > 0)
+		{
+			data->mem_location = strndup(str, len);
+			/* Convert path separators to unix type */
+			for (i = 0; i < len; ++i)
+			{
+				if (data->mem_location[i] == '\\')
+					data->mem_location[i] = '/';
+			}
+			/* Remove trailing directory separator */
+			if (data->mem_location[len-1] == '/')
+			  data->mem_location[len-1] = 0;
+		}
+		break;
+
+	case PS_FREE:
+		tmp = strndup(str, len);
+		data->mem_free = parse_ulong(tmp);
+		break;
+
+	case PS_USED:
+		tmp = strndup(str, len);
+		data->mem_used = parse_ulong(tmp);
+		break;
+
+	case PS_FILENLEN:
+		tmp = strndup(str, len);
+		data->mem_namelen = parse_ulong(tmp);
+		break;
+
+	default:
+		break;
+	}
+
+	if (tmp)
+		free(tmp);
+}
+
+static int ofs_statfs(const char *path, struct statfs *st)
 {
 	int res;
-	int size, free;
+	char *tpath = translate_path(path);
 
-	memset(st, 0, sizeof(struct statfs));
+	DEBUG("%s() %s\n", __FUNCTION__, path);
+
+	memset(st, 0, sizeof(*st));
+	st->f_bsize = 1;
+	st->f_blocks = report_space;
+	st->f_bfree = report_space;
+	st->f_bavail = report_space;
 
 	res = ofs_connect();
 	if(res < 0)
 		return res; /* errno */
 
-	/* for S45 */
-	(void) obexftp_disconnect (cli);
-	(void) obexftp_connect_uuid (cli, device, channel, UUID_S45, sizeof(UUID_S45));
- 
+#ifdef SIEMENS_S45
 	/* Retrieve Infos */
 	(void) obexftp_info(cli, 0x01);
-	size = cli->apparam_info;
+	st->f_blocks = cli->apparam_info;
 	(void) obexftp_info(cli, 0x02);
-	free = cli->apparam_info;
- 
- DEBUG("%s() GOT FS STAT: %d / %d\n", __func__, free, size);
- 
-	(void) obexftp_disconnect (cli);
-	(void) obexftp_connect (cli, device, channel);
+	st->f_bfree = cli->apparam_info;
+	st->f_bavail = st->f_bfree;
 
+#else
+	(void) obexftp_setpath(cli, tpath, 1);
+	res = obexftp_get_type(cli, XOBEX_CAPABILITY, 0, 0);
+	if (res >= 0) {
+		XML_Parser parser;
+
+		struct parsing_data data;
+		memset(&data, 0, sizeof(data));
+		data.path = tpath + 1;
+		data.path_match_len = -1;
+
+		parser = XML_ParserCreate(NULL);
+		XML_SetUserData(parser, &data);
+		XML_SetElementHandler(parser, xml_element_start, xml_element_end);
+		XML_SetCharacterDataHandler(parser, xml_character_data);
+		if (XML_Parse(parser, cli->buf_data, cli->buf_size, 1)) {
+			if (data.error) {
+				DEBUG("%s(): PARSE ERROR: %s\n", __func__, data.error);
+			} else if (data.path_match_len >= 0) {
+				*st = data.meminfo;
+			}
+		} else {
+			DEBUG("%s(): PARSE ERROR: %s at line %d\n", __func__,
+				XML_ErrorString(XML_GetErrorCode(parser)),
+				XML_GetCurrentLineNumber(parser)
+			);
+		}
+		XML_ParserFree(parser);
+	}
+#endif
+	DEBUG("%s() GOT FS STAT: %" PRId64 " / %" PRId64 "\n", __func__, st->f_bfree, st->f_blocks);
 	ofs_disconnect();
 
-	st->f_bsize = 1;	/* optimal transfer block size */
-	st->f_blocks = size;	/* total data blocks in file system */
-	st->f_bfree = free;	/* free blocks in fs */
-	st->f_bavail = free;	/* free blocks avail to non-superuser */
-
-	/* st->f_files;		/ * total file nodes in file system */
-	/* st->f_ffree;		/ * free file nodes in fs */
-	/* st->f_namelen;	/ * maximum length of filenames */
+	free(tpath);
 
 	return 0;
 }
-#endif
 
 static void *ofs_init(void) {
 	//cli_open();
@@ -465,6 +757,21 @@ static void ofs_destroy(void *UNUSED(private_data)) {
 	return;
 }
 
+static int ofs_utime(const char *a, struct utimbuf *b)
+{
+	return 0;
+}
+
+static int ofs_chmod(const char *a, mode_t b)
+{
+	return 0;
+}
+
+static int ofs_chown(const char *a, uid_t b, gid_t c)
+{
+	return 0;
+}
+
 static struct fuse_operations ofs_oper = {
 	getattr:	ofs_getattr,
 	readlink:	NULL,
@@ -479,16 +786,14 @@ static struct fuse_operations ofs_oper = {
 	rmdir:		ofs_unlink,
 	rename:		ofs_rename,
 	link:		NULL,
-	chmod:		NULL,
-	chown:		NULL,
+	chmod:		ofs_chmod,
+	chown:		ofs_chown,
 	truncate:	ofs_truncate,
-	utime:		NULL,
+	utime:		ofs_utime,
 	open:		ofs_open,
 	read:		ofs_read,
 	write:		ofs_write,
-#ifdef SIEMENS
 	statfs:		ofs_statfs,
-#endif
 	release:	ofs_release,
 	flush:		NULL,
 	fsync:		NULL,
@@ -511,13 +816,15 @@ int main(int argc, char *argv[])
 			{"usb",		required_argument, NULL, 'u'},
 			{"tty",		required_argument, NULL, 't'},
 			{"network",	required_argument, NULL, 'n'},
+			{"root",	required_argument, NULL, 'r'},
 			{"nonblock",	no_argument, NULL, 'N'},
+			{"report-space",required_argument, NULL, 'S'},
 			{"help",	no_argument, NULL, 'h'},
 			{"usage",	no_argument, NULL, 'h'},
 			{0, 0, 0, 0}
 		};
 		
-		c = getopt_long (argc, argv, "+ib:B:d:u:t:n:Nh",
+		c = getopt_long (argc, argv, "+ib:B:d:u:t:n:r:NS:h",
 				 long_options, &option_index);
 		if (c == -1)
 			break;
@@ -562,6 +869,13 @@ int main(int argc, char *argv[])
 			transport = OBEX_TRANS_INET;
 			device = optarg;
 			channel = 650;
+			{
+				char *p  =strchr(optarg, ':');
+				if (p) {
+					*p = '\0';
+					channel = atoi(++p);
+				}
+			}
 			{
                                 int n;
                                 if (sscanf(optarg, "%d.%d.%d.%d", &n, &n, &n, &n) != 4)
@@ -569,8 +883,16 @@ int main(int argc, char *argv[])
                         }
 			break;
 			
+		case 'r':
+			root = optarg;
+			break;
+
 		case 'N':
-			nonblock = 1;
+			nonblock = true;
+			break;
+
+		case 'S':
+			report_space = atoi(optarg);
 			break;
 
 		case 'h':
@@ -586,7 +908,9 @@ int main(int argc, char *argv[])
 				" -u, --usb <interface>       connect to this usb interface number\n"
 				" -t, --tty <device>          connect to this tty using a custom transport\n"
 				" -n, --network <device>      connect to this network host\n\n"
-				" -N, --nonblock              nonblocking mode\n\n"
+				" -r, --root <path>           path on device to use as root\n\n"
+				" -N, --nonblock              nonblocking mode\n"
+				" -S, --report-space <bytes>  report this number as total/free space\n\n"
 				" -h, --help, --usage         this help text\n\n"
 				"Options to fusermount need to be preceeded by two dashes (--).\n"
 				"\n",
@@ -605,6 +929,12 @@ int main(int argc, char *argv[])
 	       	fprintf(stderr, "No device selected. Use --help for help.\n");
 		exit(0);
 	}
+	if (root == NULL) root = "";
+	root_len = strlen(root);
+	while (root_len > 0 && (root[root_len-1] == '/' || root[root_len-1] == '\\')) {
+		root[root_len-1] = '\0';
+		--root_len;
+	}
 
 	argv[optind-1] = argv[0];
 
